import argparse
import asyncio

from Bing import AsyncImageGenerator


def cli_cmd():
    parser = argparse.ArgumentParser(
        prog="Bing Create",
        description="A simple lightweight AI Image Generator from text description using Bing Image Creator (DALL-E 3)",
        epilog="Made by Waenara ^^",
    )

    parser.add_argument(
        "--u", help="Your _U cookie from https://bing.com/", required=True
    )

    parser.add_argument(
        "--s", help="Your SRCHHPGUSR cookie from https://bing.com/", required=True
    )

    parser.add_argument(
        "--prompt", help="Description of image to generate", required=True
    )

    parser.add_argument(
        "--number", help="How many images to generate. Default: 4", type=int, default=4
    )

    parser.add_argument(
        "--output",
        help="Directory where to save generated images. If not specified you will just get links printed out",
    )

    parser.add_argument(
        "--quiet", help="If present logging is disabled", action="store_true"
    )

    args = parser.parse_args()
    generator = AsyncImageGenerator(
        auth_cookie_u=args.u,
        auth_cookie_srchhpgusr=args.s,
        logging_enabled=not args.quiet,
    )

    async def main():
        images = await generator.generate(prompt=args.prompt, num_images=args.number)
        if args.output:
            await generator.save(images=images, output_dir=args.output)
        else:
            print("\n".join(images))

    asyncio.run(main())



from .bing_test import (async_generator, sync_generator,
                        test_async_generate_images, test_async_save_images,
                        test_sync_generate_images, test_sync_save_images)

__all__ = [
    "sync_generator",
    "async_generator",
    "test_sync_generate_images",
    "test_sync_save_images",
    "test_async_generate_images",
    "test_async_save_images",
]



import asyncio
import json
import random
import re
import time
from typing import Dict, List

import httpx

cookies_file = "storage/cookies.json"


class CookieManager:
    def __init__(self, cookies_file: str):
        self.cookies = self._load_cookies(cookies_file)
        self.shuffled_cookies = self.cookies.copy()
        random.shuffle(self.shuffled_cookies)
        self.current_index = 0

    def _load_cookies(self, cookies_file: str) -> List[Dict[str, str]]:
        with open(cookies_file, "r") as file:
            cookies_data = json.load(file)
        return cookies_data

    def get_next_cookie(self) -> Dict[str, str]:
        if self.current_index >= len(self.shuffled_cookies):
            return None
        cookie = self.shuffled_cookies[self.current_index]
        self.current_index += 1
        return cookie

    def reset(self):
        self.shuffled_cookies = self.cookies.copy()
        random.shuffle(self.shuffled_cookies)
        self.current_index = 200


class ImageGenerator:
    def __init__(self, cookie_manager: CookieManager, logging_enabled: bool = True):
        cookie = cookie_manager.get_next_cookie()
        if not cookie:
            raise Exception("🛑 No cookies available in CookieManager!")

        self.client: httpx.Client = httpx.Client(cookies=cookie)
        self.logging_enabled = logging_enabled
        self.cookie_manager = cookie_manager

    def __log(self, message: str):
        if self.logging_enabled:
            print(message)

    def generate(self, prompt: str, num_images: int) -> list:
        images = []
        cycle = 0
        start = time.time()

        while len(images) < num_images:
            cycle += 1

            response = self.client.post(
                url=f"https://www.bing.com/images/create?q={prompt}&rt=3&FORM=GENCRE",
                data={"q": prompt, "qs": "ds"},
                follow_redirects=False,
                timeout=200,
            )

            if response.status_code == 403:
                self.__log("🛑 Cookie blocked! Resetting cookie...")
                self.client.cookies = self.cookie_manager.get_next_cookie()
                if not self.client.cookies:
                    raise Exception("🛑 No more cookies available.")
                continue

            if response.status_code != 302:
                raise Exception("🛑 Request to https://bing.com/ failed! (Redirect)")

            self.__log(f"✅ Request to https://bing.com/ sent! (cycle: {cycle})")

            if "being reviewed" in response.text or "has been blocked" in response.text:
                raise Exception("🛑 Prompt is being reviewed or blocked!")
            if "image creator in more languages" in response.text:
                raise Exception("🛑 Language is not supported by Bing yet!")

            result_id = (
                response.headers["Location"].replace("&nfy=1", "").split("id=")[-1]
            )
            results_url = f"https://www.bing.com/images/create/async/results/{result_id}?q={prompt}"

            self.__log(f"🕗 Awaiting generation... (cycle: {cycle})")
            start_time = time.time()
            while True:
                response = self.client.get(results_url)

                if time.time() - start_time > 200:
                    raise Exception("🛑 Waiting for results timed out!")

                if response.status_code != 200:
                    raise Exception(
                        "🛑 Exception happened while waiting for image generation! (NoResults)"
                    )

                if not response.text or response.text.find("errorMessage") != -1:
                    time.sleep(1)
                    continue
                else:
                    break

            new_images = [
                "https://tse" + link.split("?w=")[0]
                for link in re.findall('src="https://tse([^"]+)"', response.text)
            ]
            if len(new_images) == 0:
                raise Exception(
                    "🛑 No new images were generated for this cycle, please check your prompt"
                )
            images += new_images
            self.__log(
                f"✅ Successfully finished cycle {cycle} in {round(time.time() - start_time, 2)} seconds"
            )

        self.__log(
            f"✅ Finished generating {num_images} images in {round(time.time() - start, 2)} seconds and {cycle} cycles"
        )
        return images[:num_images]


class AsyncImageGenerator:
    def __init__(self, cookie_manager: CookieManager, logging_enabled: bool = True):
        cookie = cookie_manager.get_next_cookie()
        if not cookie:
            raise Exception("🛑 No cookies available in CookieManager!")

        self.client: httpx.AsyncClient = httpx.AsyncClient(cookies=cookie)
        self.logging_enabled = logging_enabled
        self.cookie_manager = cookie_manager

    def __log(self, message: str):
        if self.logging_enabled:
            print(message)

    async def generate(
        self, prompt: str, num_images: int, max_cycles: int = 4
    ) -> List[str]:
        images = []
        cycle = 0
        start = time.time()

        while len(images) < num_images and cycle < max_cycles:
            cycle += 1

            response = await self.client.post(
                url=f"https://www.bing.com/images/create?q={prompt}&rt=3&FORM=GENCRE",
                data={"q": prompt, "qs": "ds"},
                follow_redirects=False,
                timeout=200,
            )

            if response.status_code == 403:
                self.__log("🛑 Cookie blocked! Resetting cookie...")
                self.client.cookies = self.cookie_manager.get_next_cookie()
                if not self.client.cookies:
                    raise Exception("🛑 No more cookies available.")
                continue

            if response.status_code != 302:
                raise Exception("🛑 Request to https://bing.com/ failed! (Redirect)")

            self.__log(f"✅ Request to https://bing.com/ sent! (cycle: {cycle})")

            if "being reviewed" in response.text or "has been blocked" in response.text:
                raise Exception("🛑 Prompt is being reviewed or blocked!")
            if "image creator in more languages" in response.text:
                raise Exception("🛑 Language is not supported by Bing yet!")

            result_id = (
                response.headers["Location"].replace("&nfy=1", "").split("id=")[-1]
            )
            results_url = f"https://www.bing.com/images/create/async/results/{result_id}?q={prompt}"

            self.__log(f"🕗 Awaiting generation... (cycle: {cycle})")
            start_time = time.time()
            while True:
                response = await self.client.get(results_url)

                if time.time() - start_time > 200:
                    raise Exception("🛑 Waiting for results timed out!")

                if response.status_code != 200:
                    raise Exception(
                        "🛑 Exception happened while waiting for image generation! (NoResults)"
                    )

                if not response.text or response.text.find("errorMessage") != -1:
                    await asyncio.sleep(1)
                    continue
                else:
                    break

            new_images = [
                "https://tse" + link.split("?w=")[0]
                for link in re.findall('src="https://tse([^"]+)"', response.text)
            ]
            if len(new_images) == 0:
                raise Exception(
                    "🛑 No new images were generated for this cycle, please check your prompt"
                )
            images.extend(new_images)
            self.__log(
                f"✅ Successfully finished cycle {cycle} in {round(time.time() - start_time, 2)} seconds"
            )

        self.__log(
            f"✅ Finished generating images in {round(time.time() - start, 2)} seconds and {cycle} cycles"
        )
        return images[:num_images]



from unittest.mock import AsyncMock, Mock, patch

import pytest

from Bing.bing import AsyncImageGenerator, ImageGenerator


@pytest.fixture
def sync_generator():
    return ImageGenerator(
        auth_cookie_u="test_u", auth_cookie_srchhpgusr="test_srchhpgusr"
    )


@pytest.fixture
def async_generator():
    return AsyncImageGenerator(
        auth_cookie_u="test_u", auth_cookie_srchhpgusr="test_srchhpgusr"
    )


def test_sync_generate_images(sync_generator):
    with patch("requests.Session.post", new=Mock()) as mock_post:
        mock_post.return_value.status_code = 302
        mock_post.return_value.headers = {"Location": "https://bing.com/some_id"}
        mock_post.return_value.text = '<img src="https://tse.some_image?w=600">'

        with patch("requests.Session.get", new=Mock()) as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.text = '<img src="https://tse.some_image?w=600">'

            images = sync_generator.generate("test prompt", 1)

            assert len(images) == 1
            assert images[0] == "https://tse.some_image"


def test_sync_save_images(sync_generator):
    with patch("requests.Session.get", new=Mock()) as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.content = b"image data"

        with patch("builtins.open", new=Mock()) as mock_open:
            sync_generator.save(["https://tse.some_image"], "test_dir")

            mock_open.assert_called_once_with("test_dir/some_image.jpeg", "wb")
            mock_open().write.assert_called_once_with(b"image data")


@pytest.mark.asyncio
async def test_async_generate_images(async_generator):
    with patch("httpx.AsyncClient.post", new=AsyncMock()) as mock_post:
        mock_post.return_value.status_code = 302
        mock_post.return_value.headers = {"Location": "https://bing.com/some_id"}
        mock_post.return_value.text = '<img src="https://tse.some_image?w=600">'

        with patch("httpx.AsyncClient.get", new=AsyncMock()) as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.text = '<img src="https://tse.some_image?w=600">'

            images = await async_generator.generate("test prompt", 1)

            assert len(images) == 1
            assert images[0] == "https://tse.some_image"


@pytest.mark.asyncio
async def test_async_save_images(async_generator):
    with patch("httpx.AsyncClient.get", new=AsyncMock()) as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.content = b"image data"

        with patch("aiofiles.open", new=AsyncMock()) as mock_open:
            await async_generator.save(["https://tse.some_image"], "test_dir")

            mock_open.assert_called_once_with("test_dir/some_image.jpeg", "wb")
            mock_open().write.assert_called_once_with(b"image data")



from setuptools import find_packages, setup


try:
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
except FileNotFoundError:
    long_description = "I'm From Indonesian, and I'm still learning."

setup(
    name="gen_bing",
    version="0.0.8",
    author="Lucifer",
    author_email="ikyodeos01@gmail.com",
    description="I'm From Indonesian, and I'm still learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=["aiofiles==24.1.0", "httpx[http2]", "aiohttp>=3.9.5"],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="bing image generator gabut",
    python_requires=">=3.9",
    entry_points={
        "console_scripts": [
            "gen_bing = Bing.cli:cli_cmd",
        ],
    },
)
